import os
import sys
from optparse import SUPPRESS_HELP, OptionGroup, OptionParser

from mapdamage.rscript import check_R_lib
from mapdamage.version import __version__


def file_exist(filename):
    if os.path.exists(filename) and not os.path.isdir(filename):
        return True
    elif filename == "-":
        return True
    else:
        sys.stderr.write("Error: '%s' is not a valid file\n" % (filename))
        return None


def whereis(program):
    for path in os.environ.get("PATH", "").split(":"):
        if os.path.exists(os.path.join(path, program)) and not os.path.isdir(
            os.path.join(path, program)
        ):
            return os.path.join(path, program)
    return None


def check_py_version():
    req_version = (2, 6)
    cur_version = sys.version_info

    if cur_version >= req_version:
        return True
    else:
        sys.stderr.write(
            "Your Python interpreter is too old."
            "Please consider upgrading to at least %d.%d\n"
            % (req_version[0], req_version[1])
        )
        return None


def options():
    parser = OptionParser(
        "%prog [options] -i BAMfile -r reference.fasta\n\nUse option -h or --help for help",
        version=__version__,
        epilog="report bugs to aginolhac@snm.ku.dk, MSchubert@snm.ku.dk or jonsson.hakon@gmail.com",
    )

    args = OptionGroup(parser, "Input files")
    args.add_option(
        "-i",
        "--input",
        help="SAM/BAM file, must contain a valid header, use '-' for reading a BAM from stdin",
        action="store",
        type="string",
        dest="filename",
    )
    args.add_option(
        "-r",
        "--reference",
        help="Reference file in FASTA format",
        action="store",
        dest="ref",
    )

    parser.add_option_group(args)
    group = OptionGroup(parser, "General options")
    group.add_option(
        "-n",
        "--downsample",
        help="Downsample to a randomly selected fraction of the reads (if 0 < DOWNSAMPLE < 1), or "
        "a fixed number of randomly selected reads (if DOWNSAMPLE >= 1). By default, no downsampling is performed.",
        type=float,
        default=None,
    )
    group.add_option(
        "--downsample-seed",
        help="Seed value to use for downsampling. See documentation for py module 'random' for default behavior.",
        type=int,
        default=None,
    )
    group.add_option(
        "--merge-reference-sequences",
        help="Ignore reference sequence names when tabulating reads (using '*' instead). "
        "Useful for alignments with a large number of reference sequences, which may otherwise result in excessive "
        "memory or disk usage due to the number of tables generated.",
        default=False,
        action="store_true",
    )
    group.add_option(
        "-l",
        "--length",
        dest="length",
        help="read length, in nucleotides to consider [%default]",
        type=int,
        default=70,
        action="store",
    )
    group.add_option(
        "-a",
        "--around",
        dest="around",
        help="nucleotides to retrieve before/after reads [%default]",
        type=int,
        default=10,
        action="store",
    )
    group.add_option(
        "-Q",
        "--min-basequal",
        dest="minqual",
        help="minimum base quality Phred score considered, Phred-33 assumed [%default]",
        type=int,
        default=0,
        action="store",
    )
    group.add_option(
        "-d",
        "--folder",
        help="folder name to store results [results_FILENAME]",
        action="store",
        type="string",
        dest="folder",
    )
    group.add_option(
        "-f",
        "--fasta",
        dest="fasta",
        help="Write alignments in a FASTA file",
        default=False,
        action="store_true",
    )
    group.add_option(
        "--plot-only",
        dest="plot_only",
        help="Run only plotting from a valid result folder",
        default=False,
        action="store_true",
    )
    group.add_option(
        "-q",
        "--quiet",
        dest="quiet",
        help="Disable any output to stdout",
        default=False,
        action="store_true",
    )
    group.add_option(
        "-v",
        "--verbose",
        dest="verbose",
        help="Display progression information during parsing",
        default=False,
        action="store_true",
    )
    group.add_option(
        "--no-plot", dest="no_r", help=SUPPRESS_HELP, default=False, action="store_true"
    )
    parser.add_option_group(group)

    # options for plotting damage patterns
    group2 = OptionGroup(parser, "Options for graphics")
    group2.add_option(
        "-y",
        "--ymax",
        dest="ymax",
        help="graphical y-axis limit for nucleotide misincorporation frequencies [%default]",
        type=float,
        default=0.3,
        action="store",
    )
    group2.add_option(
        "-m",
        "--readplot",
        dest="readplot",
        help="read length, in nucleotides, considered for plotting nucleotide misincorporations [%default]",
        type=int,
        default=25,
        action="store",
    )
    group2.add_option(
        "-b",
        "--refplot",
        dest="refplot",
        help="the number of reference nucleotides to consider for plotting base composition in the region located upstream "
        "and downstream of every read [%default]",
        type=int,
        default=10,
        action="store",
    )
    group2.add_option(
        "-t",
        "--title",
        dest="title",
        help="title used for plots [%default]",
        type="string",
        default="",
        action="store",
    )
    parser.add_option_group(group2)

    # Then the plethora of optional options for the statistical estimation ..
    group3 = OptionGroup(parser, "Options for the statistical estimation")
    group3.add_option(
        "",
        "--rand",
        dest="rand",
        help="Number of random starting points for the likelihood optimization  [%default]",
        type=int,
        default=30,
        action="store",
    )
    group3.add_option(
        "",
        "--burn",
        dest="burn",
        help="Number of burn-in iterations  [%default]",
        type=int,
        default=10000,
        action="store",
    )
    group3.add_option(
        "",
        "--adjust",
        dest="adjust",
        help="Number of adjust proposal variance parameters iterations  [%default]",
        type=int,
        default=10,
        action="store",
    )
    group3.add_option(
        "",
        "--iter",
        dest="iter",
        help="Number of final MCMC iterations  [%default]",
        type=int,
        default=50000,
        action="store",
    )
    group3.add_option(
        "",
        "--forward",
        dest="forward",
        help="Using only the 5' end of the seqs  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--reverse",
        dest="reverse",
        help="Using only the 3' end of the seqs  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--var-disp",
        dest="var_disp",
        help="Variable dispersion in the overhangs  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--jukes-cantor",
        dest="jukes_cantor",
        help="Use Jukes Cantor instead of HKY85  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--diff-hangs",
        dest="diff_hangs",
        help="The overhangs are different for 5' and 3'  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--fix-nicks",
        dest="fix_nicks",
        help="Fix the nick frequency vector (Only C.T from the 5' end and G.A from the 3' end)  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--use-raw-nick-freq",
        dest="use_raw_nick_freq",
        help="Use the raw nick frequency vector without smoothing  [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--single-stranded",
        dest="single_stranded",
        help="Single stranded protocol [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--theme-bw",
        dest="theme_bw",
        help="Use black and white theme in post. pred. plot [%default]",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "",
        "--seq-length",
        dest="seq_length",
        help="How long sequence to use from each side [%default]",
        type=int,
        default=12,
        action="store",
    )
    group3.add_option(
        "--stats-only",
        dest="stats_only",
        help="Run only statistical estimation from a valid result folder",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "--no-stats",
        help="Disabled statistical estimation, active by default",
        default=False,
        action="store_true",
    )
    group3.add_option(
        "--check-R-packages",
        help="Check if the R modules are working",
        default=False,
        action="store_true",
    )
    parser.add_option_group(group3)

    group4 = OptionGroup(parser, "Options for rescaling of BAM files")
    group4.add_option(
        "--rescale",
        dest="rescale",
        help="Rescale the quality scores in the BAM file using the output from the statistical estimation",
        default=False,
        action="store_true",
    )
    group4.add_option(
        "--rescale-only",
        dest="rescale_only",
        help="Run only rescaling from a valid result folder",
        default=False,
        action="store_true",
    )
    group4.add_option(
        "--rescale-out",
        dest="rescale_out",
        help="Write the rescaled BAM to this file",
        default=None,
        action="store",
    )
    group4.add_option(
        "--rescale-length-5p",
        dest="rescale_length_5p",
        help="How many bases to rescale at the 5' termini; defaults to --seq-length.",
        type=int,
        action="store",
    )
    group4.add_option(
        "--rescale-length-3p",
        dest="rescale_length_3p",
        help="How many bases to rescale at the 5' termini; defaults to --seq-length.",
        type=int,
        action="store",
    )
    parser.add_option_group(group4)

    # Parse the arguments
    (options, args) = parser.parse_args()

    # check python version
    if not check_py_version():
        return None

    # if the user wants to check the R packages then do that before the option parsing
    if options.check_R_packages:
        if check_R_lib():
            sys.exit(1)
        else:
            print("All R packages are present")
            sys.exit(0)

    # check general arguments
    if not (options.plot_only or options.stats_only) and not options.filename:
        parser.error("SAM/BAM file not given (-i)")
    if not (options.plot_only or options.ref):
        parser.error("Reference file not given (-r)")
    if not options.plot_only and not options.stats_only:
        if not file_exist(options.filename) or not file_exist(options.ref):
            return None
    if options.downsample is not None:
        if options.downsample <= 0:
            parser.error("-n/--downsample must be a positive value")
        elif options.downsample >= 1:
            options.downsample = int(options.downsample)

    if options.plot_only and not options.folder:
        parser.error("Folder not provided, required with --plot-only")
    if options.stats_only and not options.folder:
        parser.error("Folder not provided, required with --stats-only")
    if options.rescale_only and not options.folder:
        parser.error("Folder not provided, required with --rescale-only")
    if options.rescale_only and not options.filename:
        parser.error("Input bam not provided, required with --rescale-only")
    if options.rescale_only and not options.ref:
        parser.error("Reference not provided, required with --rescale-only")

    if options.verbose and options.quiet:
        parser.error("Cannot use verbose and quiet option at the same time")

    # check options
    if options.length < 0:
        parser.error("length (-l) must be a positive integrer")
    if options.around < 0:
        parser.error("around (-a) must be a positive integrer")
    if options.ymax <= 0 or options.ymax > 1:
        parser.error("ymax (-b) must be an real number between 0 and 1")
    if options.readplot < 0:
        parser.error("readplot (-m) must be a positive integrer")
    if options.refplot < 0:
        parser.error("refplot (-b) must be a positive integrer")
    if options.refplot > options.around and not options.plot_only:
        parser.error("refplot (-b) must be inferior to around (-a)")
    if options.readplot > options.length:
        parser.error("readplot (-m) must be inferior to length (-l)")
    if options.minqual < 0 or options.minqual > 41:
        parser.error(
            "minimal base quality, Phred score, must be within this range: 0 - 41"
        )

    # check statistic options
    if options.forward and options.reverse:
        parser.error(
            "Cannot use only forward end and only reverse end for the statistics"
        )

    # use filename as default for plot titles if not set
    if options.title == "" and options.filename:
        options.title = os.path.splitext(os.path.basename(options.filename))[0]
    # for --plot-only, use the folder name, without results_ as title
    if options.title == "" and not options.filename and options.folder:
        options.title = os.path.splitext(os.path.basename(options.folder))[0].replace(
            "results_", ""
        )

    # check folder
    if not options.folder and options.filename:
        options.folder = (
            "results_" + os.path.splitext(os.path.basename(options.filename))[0]
        )

    # check destination for rescaled bam
    if not options.rescale_out and (options.rescale or options.rescale_only):
        # if there are multiple bam files to rescale then pick first one as
        # the name of the rescaled file
        if isinstance(options.filename, list):
            basename = os.path.basename(options.filename[0])
        else:
            basename = os.path.basename(options.filename)
        with_ext = os.path.splitext(basename)[0] + ".rescaled.bam"
        options.rescale_out = os.path.join(options.folder, with_ext)

    if os.path.isdir(options.folder):
        if not options.quiet and not options.plot_only:
            print(("Warning, %s already exists" % options.folder))
        if options.plot_only:
            if not file_exist(options.folder + "/dnacomp.txt") or not file_exist(
                options.folder + "/misincorporation.txt"
            ):
                parser.error("folder %s is not a valid result folder" % options.folder)
    else:
        os.makedirs(options.folder, mode=0o750)
        if options.plot_only or options.stats_only or options.rescale_only:
            sys.stderr.write(
                "Error, %s does not exist while plot/stats/rescale only was used\n"
                % options.folder
            )
            return None

    if options.rescale_length_3p is None:
        options.rescale_length_3p = options.seq_length
    elif not (0 <= options.rescale_length_3p <= options.seq_length):
        parser.error(
            "--rescale-length-3p must be less than or equal to "
            "--seq-length and greater than zero"
        )

    if options.rescale_length_5p is None:
        options.rescale_length_5p = options.seq_length
    elif not (0 <= options.rescale_length_5p <= options.seq_length):
        parser.error(
            "--rescale-length-5p must be less than or equal to "
            "--seq-length and greater than zero"
        )

    # check if the Rscript executable is present on the system
    if not whereis("Rscript"):
        print("Warning, Rscript is not in your PATH, plotting is disabled")
        options.no_r = True

    # check the nick frequencies options
    if (options.use_raw_nick_freq + options.fix_nicks + options.single_stranded) > 1:
        parser.error(
            "The options --use-raw-nick-freq, --fix-nicks and --single-stranded are mutually exclusive."
        )

    if check_R_lib():
        # check for R libraries
        print("The Bayesian estimation has been disabled\n")
        options.no_stats = True
        if options.stats_only:
            sys.exit("Cannot use --stats-only with missing R libraries")
        if options.rescale:
            sys.exit("Cannot use --rescale with missing R libraries")
        if options.rescale_only:
            sys.exit("Cannot use --rescale-only with missing R libraries")

    return options



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
import random
import sys
import time

import pysam

import mapdamage

""" Copyright (c) 2012  Aurélien Ginolhac, Mikkel Schubert, Hákon Jónsson
and Ludovic Orlando

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
OR OTHER DEALINGS IN THE SOFTWARE.

plot and quantify damage patterns from a SAM/BAM file

:Authors: Aurélien Ginolhac, Mikkel Schubert, Hákon Jónsson, Ludovic Orlando
:Contact: aginolhac@snm.ku.dk, MSchubert@snm.ku.dk, jonsson.hakon@gmail.com
:Date: November 2012
:Type: tool
:Input: SAM/BAM
:Output: tabulated tables, pdf
"""

_BAM_UNMAPPED = 0x4
_BAM_SECONDARY = 0x100
_BAM_FAILED_QC = 0x200
_BAM_PCR_DUPE = 0x400
_BAM_CHIMERIC = 0x800


def _filter_reads(bamfile):
    filtered_flags = (
        _BAM_UNMAPPED | _BAM_SECONDARY | _BAM_FAILED_QC | _BAM_PCR_DUPE | _BAM_CHIMERIC
    )

    for read in bamfile:
        if not (read.flag & filtered_flags):
            yield read


def _downsample_to_fraction(bamfile, options):
    if options.verbose:
        print("\tDownsampling %.2f fraction of the original file" % options.downsample)
    assert options.downsample > 0, (
        "Downsample fraction must be positive, not %s" % options.downsample
    )

    rand = random.Random(options.downsample_seed)
    for read in _filter_reads(bamfile):
        if rand.random() < options.downsample:
            yield read


def _downsample_to_fixed_number(bamfile, options):
    if options.verbose:
        print(
            "\tDownsampling %d random reads from the original file" % options.downsample
        )

    # use reservoir sampling
    downsample_to = int(options.downsample)
    rand = random.Random(options.downsample_seed)
    sample = [None] * downsample_to
    for index, record in enumerate(_filter_reads(bamfile)):
        if index >= downsample_to:
            index = rand.randint(0, index)
            if index >= downsample_to:
                continue
        sample[index] = record
    return [_f for _f in sample if _f]


def _read_bamfile(bamfile, options):
    """
    Takes a subset of the bamfile. Can use a approximate fraction of the
    hits or specific number of reads using reservoir sampling. Returns
    a list in the last case otherwise a iterator.
    """
    if options.downsample is None:
        return _filter_reads(bamfile)
    elif options.downsample < 1:
        return _downsample_to_fraction(bamfile, options)
    else:
        return _downsample_to_fixed_number(bamfile, options)


def main():
    start_time = time.time()

    options = mapdamage.parseoptions.options()
    if options is None:
        sys.stderr.write("Option parsing failed, terminating the program\n")
        return 1

    logging.basicConfig(
        filename=os.path.join(options.folder, "Runtime_log.txt"),
        format="%(asctime)s\t%(levelname)s\t%(name)s: %(message)s",
        level=logging.DEBUG,
    )
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    logging.getLogger().addHandler(handler)

    logger = logging.getLogger("main")
    logger.info("Started with the command: " + " ".join(sys.argv))

    # plot using R if results folder already done
    if options.plot_only:
        if options.no_r:
            logger.error(
                "Cannot use plot damage patterns if R is missing, terminating the program"
            )
            return 1
        else:
            mapdamage.rscript.plot(options)
            mapdamage.rscript.opt_plots(options)
            return 0

    # run the Bayesian estimation if the matrix construction is done
    if options.stats_only:
        # does not work for very low damage levels
        if mapdamage.tables.check_table_and_warn_if_dmg_freq_is_low(options.folder):
            logger.error("Cannot use the Bayesian estimation, terminating the program")
            return 1
        else:
            # before running the Bayesian estimation get the base composition
            path_to_basecomp = os.path.join(options.folder, "dnacomp_genome.csv")
            if os.path.isfile(path_to_basecomp):
                # Try to read the base composition file
                mapdamage.composition.read_base_comp(path_to_basecomp)
            else:
                # Construct the base composition file
                mapdamage.composition.write_base_comp(options.ref, path_to_basecomp)
            mapdamage.rscript.run_stats(options)
            return 0

    # fetch all references and associated lengths in nucleotides
    try:
        ref = pysam.Fastafile(options.ref)
    except IOError as e:
        # Couldn't open the reference file
        logger.error(
            "Couldn't open the reference file "
            + options.ref
            + " or there are permission problems writing the "
            + options.ref
            + ".fai file"
        )
        raise e

    # rescale the qualities
    if options.rescale_only:
        if not options.quiet:
            print("Starting rescaling...")
        mapdamage.rescale.rescale_qual(ref, options)
        return 0

    # open SAM/BAM file
    if options.filename == "-":
        in_bam = pysam.Samfile("-", "rb")
        # disabling rescaling if reading from standard input since we need
        # to read it twice
        if options.rescale:
            print("Warning, reading from standard input, rescaling is disabled")
            options.rescale = False
    else:
        in_bam = pysam.Samfile(options.filename)

    reflengths = dict(list(zip(in_bam.references, in_bam.lengths)))
    # check if references in SAM/BAM are the same in the fasta reference file
    fai_lengths = mapdamage.seq.read_fasta_index(options.ref + ".fai")
    if not fai_lengths:
        return 1
    elif not mapdamage.seq.compare_sequence_dicts(fai_lengths, reflengths):
        return 1
    elif (len(reflengths) >= 1000) and not options.merge_reference_sequences:
        print(
            "WARNING: Alignment contains a large number of reference sequences (%i)!"
            % len(reflengths)
        )
        print("  This may lead to excessive memory/disk usage.")
        print("  Consider using --merge-reference-sequences")
        print()

    refnames = in_bam.references
    if options.merge_reference_sequences:
        refnames = ["*"]

    # for misincorporation patterns, record mismatches
    misincorp = mapdamage.tables.initialize_mut(refnames, options.length)
    # for fragmentation patterns, record base compositions
    dnacomp = mapdamage.tables.initialize_comp(refnames, options.around, options.length)
    # for length distributions
    lgdistrib = mapdamage.tables.initialize_lg()

    if not options.quiet:
        print("\tReading from '%s'" % options.filename)
        if options.minqual != 0:
            print("\tFiltering out bases with a Phred score < %d" % options.minqual)
        if options.verbose:
            print(
                "\t%d references are assumed in SAM/BAM file, for a total of %d nucleotides"
                % (len(reflengths), sum(reflengths.values()))
            )
        print("\tWriting results to '%s/'" % options.folder)

    # open file handler to write alignments in fasta format
    if options.fasta:
        # use name of the SAM/BAM filename without extension
        ffasta = os.path.splitext(os.path.basename(options.filename))[0] + ".fasta"
        if not options.quiet:
            print("\tWriting alignments in '%s'" % ffasta)
        fhfasta = open(options.folder + "/" + ffasta, "w")

    counter = 0

    # main loop
    for read in _read_bamfile(in_bam, options):
        counter += 1

        # external coordinates 5' and 3' , 3' is 1-based offset
        coordinate = mapdamage.align.get_coordinates(read)
        # record aligned length for single-end reads
        lgdistrib = mapdamage.seq.record_lg(read, coordinate, lgdistrib)
        # fetch reference name, chromosome or contig names
        chrom = in_bam.getrname(read.tid)

        (before, after) = mapdamage.align.get_around(
            coordinate, chrom, reflengths, options.around, ref
        )
        refseq = ref.fetch(chrom, min(coordinate), max(coordinate)).upper()
        # read.query contains aligned sequences while read.seq is the read itself
        seq = read.query

        # add gaps according to the cigar string, do it for qualities if filtering options is on
        if not (options.minqual and read.qual):
            if options.minqual:
                logger.warning(
                    "Cannot filter bases by PHRED scores for read '%s'; no scores available."
                    % read.qname
                )

            (seq, refseq) = mapdamage.align.align(read.cigar, seq, refseq)
        else:
            # add gaps to qualities and mask read and reference nucleotides if below desired threshold
            (seq, _, refseq) = mapdamage.align.align_with_qual(
                read.cigar, seq, read.qqual, options.minqual, refseq
            )

        # reverse complement read and reference when mapped reverse strand
        if read.is_reverse:
            refseq = mapdamage.seq.revcomp(refseq)
            seq = mapdamage.seq.revcomp(seq)
            beforerev = mapdamage.seq.revcomp(after)
            after = mapdamage.seq.revcomp(before)
            before = beforerev

        if options.merge_reference_sequences:
            chrom = "*"

        # record soft clipping when present
        mapdamage.align.record_soft_clipping(read, misincorp[chrom], options.length)

        # count misincorparations by comparing read and reference base by base
        mapdamage.align.get_mis(
            read, seq, refseq, chrom, options.length, misincorp, "5p"
        )
        # do the same with sequences align to 3'-ends
        mapdamage.align.get_mis(
            read, seq[::-1], refseq[::-1], chrom, options.length, misincorp, "3p"
        )
        # compute base composition for reads
        mapdamage.composition.count_read_comp(read, chrom, options.length, dnacomp)

        # compute base composition for genomic regions
        mapdamage.composition.count_ref_comp(read, chrom, before, after, dnacomp)

        if options.fasta:
            mapdamage.seq.write_fasta(
                read,
                chrom,
                seq,
                refseq,
                min(coordinate),
                max(coordinate),
                before,
                after,
                fhfasta,
            )

        if options.verbose:
            if counter % 50000 == 0:
                print("\t%10d filtered alignments processed" % (counter,))

    if options.verbose:
        print("\tDone. %d filtered alignments processed" % (counter,))
    logger.debug("BAM read in %f seconds" % (time.time() - start_time,))

    # close file handlers
    in_bam.close()
    if options.fasta:
        fhfasta.close()

    # output results, write summary tables to disk
    with open(options.folder + "/" + "misincorporation.txt", "w") as fmut:
        mapdamage.tables.print_mut(misincorp, options, fmut)
    with open(options.folder + "/" + "dnacomp.txt", "w") as fcomp:
        mapdamage.tables.print_comp(dnacomp, options, fcomp)
    with open(options.folder + "/" + "lgdistribution.txt", "w") as flg:
        mapdamage.tables.print_lg(lgdistrib, options, flg)

    # plot using R
    if not options.no_r:
        mapdamage.rscript.plot(options)
        mapdamage.rscript.opt_plots(options)

    # raises a warning for very low damage levels
    if mapdamage.tables.check_table_and_warn_if_dmg_freq_is_low(options.folder):
        options.no_stats = True

    # run the Bayesian estimation
    if not options.no_stats:
        # before running the Bayesian estimation get the base composition
        mapdamage.composition.write_base_comp(
            options.ref, os.path.join(options.folder, "dnacomp_genome.csv")
        )
        mapdamage.rscript.run_stats(options)

    # rescale the qualities
    if options.rescale:
        mapdamage.rescale.rescale_qual(ref, options)

    # need the fasta reference still open for rescaling
    ref.close()

    # log the time it took
    logger.info("Successful run")
    logger.debug("Run completed in %f seconds" % (time.time() - start_time,))

    return 0



import filecmp
import optparse
import subprocess
import unittest

import pysam

import mapdamage


def read_nocomment(dnacomp):
    with open(dnacomp, "r") as f:
        a = f.readlines()
        return [x for x in a if not x.startswith("#")]


def mock_options(filename, rescale_out, folder):
    """Make the options object with nice values for testing"""
    return optparse.Values(
        {
            "filename": filename,
            "rescale_out": rescale_out,
            "verbose": True,
            "folder": folder,
            "rescale_length_5p": 12,  # default values as in --seq-length
            "rescale_length_3p": 12,  # default values as in --seq-length
            "quiet": True,
        }
    )


class testCases(unittest.TestCase):
    def test_no_stats(self):
        """regular mapDamage"""
        subprocess.run(
            [
                "mapDamage",
                "-i",
                "tests/test.bam",
                "-r",
                "tests/fake1.fasta",
                "-d",
                "tests/results",
                "--no-stats",
            ],
            check=True,
        )
        self.assertTrue(
            read_nocomment("tests/dnacomp.txt")
            == read_nocomment("tests/results/dnacomp.txt")
        )


class testRescaling(unittest.TestCase):
    def test_single_end_file(self):
        """Test, rescaling BAM file"""
        #
        # The expected substitution frequencies before and after scaling using the scaled qualities as probalities:
        # CT	0.06226411977920493		0.04163524443356556
        # TC	0.020395286584806528		0.020395286584806528
        # GA	0.04400459948304954		0.03794905109091021
        # AG	0.05355350777642983		0.05355350777642983
        # Quality metrics before and after scaling
        # CT-Q0 	5		5
        # CT-Q10 	5		2
        # CT-Q20 	3		2
        # CT-Q30 	3		2
        # CT-Q40 	0		0
        # GA-Q0 	5		5
        # GA-Q10 	5		4
        # GA-Q20 	1		0
        # GA-Q30 	1		0
        # GA-Q40 	0		0
        options = mock_options(
            "tests/test.bam", "tests/test.rescaled.sam", "tests/probs/"
        )
        ref = pysam.Fastafile("tests/fake1.fasta")
        mapdamage.rescale.rescale_qual(ref, options, debug=True)
        self.assertTrue(
            filecmp.cmp("tests/test.rescaled.sam", "tests/test.rescaled.correct.sam")
        )


if __name__ == "__main__":
    unittest.main()



import collections
import os

import mapdamage
from mapdamage.version import __version__


def initialize_mut(ref, length):
    tab = {}
    for contig in ref:
        tab_contig = tab[contig] = {}
        for end in ("5p", "3p"):
            tab_end = tab_contig[end] = {}
            for std in ("+", "-"):
                tab_std = tab_end[std] = {}
                for mut in mapdamage.seq.HEADER:
                    tab_std[mut] = dict.fromkeys(range(length), 0)

    return tab


def print_mut(mut, opt, out):
    _print_freq_table(mut, mapdamage.seq.HEADER, opt, out, offset=1)


def initialize_comp(ref, around, length):
    keys = {
        "3p": list(range(-length, 0)) + list(range(1, around + 1)),
        "5p": list(range(-around, 0)) + list(range(1, length + 1)),
    }

    tab = {}
    for contig in ref:
        tab_contig = tab[contig] = {}
        for end in ("5p", "3p"):
            tab_end = tab_contig[end] = {}
            for std in ("+", "-"):
                tab_std = tab_end[std] = {}
                for letters in mapdamage.seq.LETTERS:
                    tab_std[letters] = dict.fromkeys(keys[end], 0)

    return tab


def print_comp(comp, opt, out):
    columns = mapdamage.seq.LETTERS + ("Total",)
    _print_freq_table(comp, columns, opt, out)


def initialize_lg():
    tab = {}
    for std in ("+", "-"):
        tab[std] = collections.defaultdict(int)

    return tab


def print_lg(tab, opt, out):
    out.write("# table produced by mapDamage version %s\n" % __version__)
    out.write(
        "# using mapped file %s and %s as reference file\n" % (opt.filename, opt.ref)
    )
    if opt.minqual != 0:
        out.write(
            "# Quality filtering of bases with a Phred score < %d\n" % opt.minqual
        )
    out.write("# Std: strand of reads\n")
    out.write("Std\tLength\tOccurences \n")
    for std in tab:
        for i in tab[std]:
            out.write("%s\t%d\t%d\n" % (std, i, tab[std][i]))


def check_table_and_warn_if_dmg_freq_is_low(folder):
    """Returns true if the damage frequencies are too low to allow
    Bayesian estimation of DNA damages, i.e < 1% at first position.
    """
    total = 0.0
    for filename in ("5pCtoT_freq.txt", "3pGtoA_freq.txt"):
        if not os.path.exists(folder + "/" + filename):
            print(
                "Error: Required table has not been created ('%s'), bayesian computation cannot be performed"
                % filename
            )
            return True

        with open(os.path.join(folder, filename)) as handle:
            for line in handle:
                freq = line.strip().split("\t")
                if freq[0] == "1":
                    total += float(freq[1])
                    break
            else:
                print(
                    "Error: Could not find pos = 1 in table '%s', bayesian computation cannot be performed"
                    % filename
                )
                return True

    if total < 0.01:
        print(
            "Warning: DNA damage levels are too low, the Bayesian computation should not be performed (%f < 0.01)\n"
            % total
        )

    return False


def _print_freq_table(table, columns, opt, out, offset=0):
    out.write("# table produced by mapDamage version %s\n" % __version__)
    out.write(
        "# using mapped file %s and %s as reference file\n" % (opt.filename, opt.ref)
    )
    if opt.minqual != 0:
        out.write("# Bases with a Phred score < %d were filtered out\n" % opt.minqual)
    out.write(
        "# Chr: reference from sam/bam header, End: from which termini of DNA sequences, Std: strand of reads\n"
    )
    out.write("Chr\tEnd\tStd\tPos\t%s\n" % ("\t".join(columns)))

    for reference, ends in sorted(table.items()):
        for end, strands in sorted(ends.items()):
            for strand, subtable in sorted(strands.items()):
                subtable["Total"] = {}
                for index in sorted(subtable[columns[0]]):
                    subtable["Total"][index] = sum(
                        subtable[letter][index] for letter in mapdamage.seq.LETTERS
                    )

                    out.write(
                        "%s\t%s\t%s\t%d" % (reference, end, strand, index + offset)
                    )
                    for base in columns:
                        out.write("\t%d" % subtable[base][index])
                    out.write("\n")



import csv

import mapdamage
import mapdamage.seqtk


def count_ref_comp(read, chrom, before, after, comp):
    """record basae composition in external genomic regions"""
    std = "-" if read.is_reverse else "+"

    _update_table(comp[chrom]["5p"][std], before, range(-len(before), 0))
    _update_table(comp[chrom]["3p"][std], after, range(1, len(after) + 1))


def count_read_comp(read, chrom, length, comp):
    """record base composition of read, discard marked nucleotides"""
    std, seq = "+", read.query
    if read.is_reverse:
        std, seq = "-", mapdamage.seq.revcomp(seq)

    _update_table(comp[chrom]["5p"][std], seq, range(1, length + 1))
    _update_table(comp[chrom]["3p"][std], reversed(seq), range(-1, -length - 1, -1))


def _update_table(table, sequence, indices):
    for index, nt in zip(indices, sequence):
        if nt in table:
            table[nt][index] += 1


def write_base_comp(fasta, destination):
    """Calculates the total base composition across all sequences in 'fasta'
    and writes them to 'destination' as CSV.
    """
    bases = {"A": 0, "C": 0, "G": 0, "T": 0}
    for stats in mapdamage.seqtk.comp(str(fasta)):
        for key in bases:
            bases[key] += stats[key]

    # calculate the base frequencies
    ba_su = sum(bases.values())
    for key in bases:
        bases[key] = bases[key] / ba_su

    with open(destination, "wt", newline="") as handle:
        writer = csv.writer(handle)

        header = ["A", "C", "G", "T"]
        writer.writerow(header)
        writer.writerow(bases[key] for key in header)


def read_base_comp(filename):
    """Read the base compition from a file created by write_base_comp"""
    with open(filename, newline="") as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            return row

    raise csv.Error("No rows found in %r" % (filename,))



#!/usr/bin/env python3
from setuptools import Extension, setup

setup(
    ext_modules=[
        Extension(
            "mapdamage.seqtk",
            sources=["mapdamage/seqtk/seqtk.c"],
            libraries=["z"],
        )
    ],
)



import sys

# from Martin Kircher, to complement DNA
TABLE = str.maketrans("TGCAMRWSYKVHDBtgcamrwsykvhdb", "ACGTKYWSRMBDHVacgtkywsrmbdhv")

LETTERS = ("A", "C", "G", "T")
MUTATIONS = (
    "G>A",
    "C>T",
    "A>G",
    "T>C",
    "A>C",
    "A>T",
    "C>G",
    "C>A",
    "T>G",
    "T>A",
    "G>C",
    "G>T",
    "A>-",
    "T>-",
    "C>-",
    "G>-",
    "->A",
    "->T",
    "->C",
    "->G",
    "S",
)
HEADER = LETTERS + ("Total",) + MUTATIONS


def write_fasta(read, ref, seq, refseq, start, end, before, after, fout):
    std = "-" if read.is_reverse else "+"

    # output coordinate in 1-based offset for 5'
    if len(before) > 0:
        fout.write(">%s\n%s\n" % (ref, before))
    fout.write(
        ">%s:%d-%d\n%s\n>%s_%s\n%s\n"
        % (ref, start + 1, end, refseq, read.qname, std, seq)
    )
    if len(after) > 0:
        fout.write(">%s\n%s\n" % (ref, after))


def revcomp(seq):
    """return reverse complemented string"""
    return seq.translate(TABLE)[::-1]


def record_lg(read, coordinate, tab):
    """record global length distribution
    don't record paired reads as they are normally not used for aDNA"""
    std = "-" if read.is_reverse else "+"

    length = max(coordinate) - min(coordinate)
    if not read.is_paired:
        tab[std][length] = tab[std][length] + 1

    return tab


def read_fasta_index(filename):
    """from a fasta index file, fai, return dictionary of references:lengths"""

    def print_err(msg, filename, line):
        sys.stderr.write("Error: %s\n" % msg)
        sys.stderr.write("       Filename: %s\n" % filename)
        sys.stderr.write("       Line:     %s\n" % repr(line))

    fai = {}
    with open(filename, "r") as handle:
        for line in handle:
            ref = line.split("\t")
            if len(ref) != 5:
                print_err(
                    "Line in fasta index contains wrong number of fields, found %i, expected 5:"
                    % len(ref),
                    filename,
                    line,
                )
                return None

            try:
                fai[ref[0]] = int(ref[1])
            except ValueError:
                print_err(
                    "Column 2 in FASTA index did not contain a number, found '%s':"
                    % ref[1],
                    filename,
                    line,
                )
                return None

    if not fai:
        sys.stderr.write(
            "Error: Index for %r does contain any sequences.\n" % (filename,)
        )
        sys.stderr.write("       Please ensure that FASTA file is valid, and\n")
        sys.stderr.write("       re-index file using 'samtool faidx'.\n")
        return None

    return fai


def compare_sequence_dicts(fasta_dict, bam_dict):
    """Compares a FASTA and BAM sequence dictionary, and prints any differences.
    Returns true if all required sequences are found, false otherwise."""
    if fasta_dict == bam_dict:
        return True

    sys.stderr.write("Sequence dictionaries in FASTA/BAM files differ:\n")
    common = set(fasta_dict) & set(bam_dict)
    if not common:
        sys.stderr.write("FATAL ERROR: No sequences in common!\n")
        return False

    # Check that the lengths of required sequences match (fatal error)
    different = []
    for key in sorted(common):
        if fasta_dict[key] != bam_dict[key]:
            different.append((key, fasta_dict[key], bam_dict[key]))

    if different:
        sys.stderr.write("FATAL ERROR: Length of required sequences differ:\n")
        for values in different:
            sys.stderr.write("    - %s: %i bp vs %i bp\n" % values)

    # Check for sequences only found in the BAM file (fatal errors)
    bam_only = set(bam_dict) - common
    if bam_only:
        sys.stderr.write("FATAL ERROR: Sequences missing from FASTA dictionary:\n")
        for key in bam_only:
            sys.stderr.write("    - %s = %i bp\n" % (key, bam_dict[key]))

    # Check for sequences only found in the BAM file (fatal errors)
    fasta_only = set(fasta_dict) - common
    if fasta_only:
        sys.stderr.write("WARNING: FASTA file contains extra sequences:\n")
        for key in fasta_only:
            sys.stderr.write("    - %s = %i bp\n" % (key, fasta_dict[key]))

    sys.stderr.write("\n")

    return not (different or bam_only)



import logging
import os
import subprocess
import time
from subprocess import CalledProcessError, check_call

import mapdamage
from mapdamage.version import __version__


def construct_path(name, folder="Rscripts"):
    """Construct a path to the mapdamage package data given the name"""
    return os.path.join(mapdamage.__path__[0], folder, name)


def plot(opt):
    """
    Calls the R script to make the plots, takes in the arguments
    from parameter processing
    """
    # comp<-args[1]
    # pdfout<-args[2]
    # around<-as.numeric(args[3])
    # misincorp<-args[4]
    # lg<-as.numeric(args[5])
    # ymax<-as.numeric(args[6])
    # folder<-args[7]
    # title<-args[8]
    # version<-args[9]

    fmut = opt.folder + "/" + "misincorporation.txt"
    fcomp = opt.folder + "/" + "dnacomp.txt"
    title = opt.folder + "/" + "Fragmisincorporation_plot.pdf"

    script = construct_path("mapDamage.R")
    call = [
        "Rscript",
        script,
        fcomp,
        title,
        opt.refplot,
        fmut,
        opt.readplot,
        opt.ymax,
        opt.folder,
        opt.title,
        __version__,
    ]
    code = subprocess.call(list(map(str, call)))

    if code == 0:
        if not opt.quiet:
            print(("pdf %s generated" % title))
        return 0
    else:
        print("Error: plotting with R failed")
        return 1


def opt_plots(opt):
    """optional plots for length distribution
    and cumulative C>T mutations, per strand"""

    fmut = opt.folder + "/" + "misincorporation.txt"
    flength = opt.folder + "/" + "lgdistribution.txt"
    output = opt.folder + "/" + "Length_plot.pdf"

    script = construct_path("lengths.R")
    call = [
        "Rscript",
        script,
        flength,
        output,
        fmut,
        opt.length,
        opt.title,
        __version__,
        bool_to_int(opt.quiet),
    ]
    code = subprocess.call(list(map(str, call)))
    if code == 0:
        return 0
    else:
        print("Error: plotting with R failed")
        return 1


def check_R_one_lib(name):
    """Checks if a necessary R library is here."""
    with open(os.devnull, "w") as devnull:
        rpa = construct_path("stats/checkLibraries.R")
        return subprocess.call(
            ["Rscript", rpa, "--args", name], stdout=devnull, stderr=devnull
        )


def check_R_lib():
    """
    Checks if the necessary R libraries are here, signal
    otherwise
    """
    libs = ["inline", "ggplot2", "gam", "Rcpp", "RcppGSL"]
    missing_lib = []
    for lib in libs:
        # Check the libraries
        if check_R_one_lib(lib):
            # found a missing library
            missing_lib.append(lib)
    if len(missing_lib) > 1:
        # Grammar Nazi has arrived
        last_ele = missing_lib.pop()
        print(
            (
                "Missing the following R libraries '"
                + "', '".join(missing_lib)
                + "' and '"
                + last_ele
                + "'"
            )
        )
        return 1
    elif len(missing_lib) == 1:
        print(("Missing the following R library " + missing_lib[0]))
        return 1
    else:
        # No missing libraries
        return 0


def run_stats(opt):
    """
    Runs the Bayesian estimation program, using the options opt
    """
    arg = [
        "Rscript",
        construct_path("stats/runGeneral.R"),
        "--args",
        opt.rand,
        opt.burn,
        opt.adjust,
        opt.iter,
        bool_to_int(opt.forward),
        bool_to_int(opt.reverse),
        bool_to_int(not opt.var_disp),
        bool_to_int(not opt.diff_hangs),
        0,
        bool_to_int(opt.fix_nicks),
        bool_to_int(not opt.single_stranded),
        opt.seq_length,
        opt.folder + "/",
        construct_path("stats/"),
        opt.folder + "/Stats_out",
        int(opt.verbose),
        int(opt.quiet),
        int(opt.jukes_cantor),
        opt.folder + "/acgt_ratio.csv",
        int(opt.use_raw_nick_freq),
        int(opt.theme_bw),
    ]
    arg = [str(i) for i in arg]

    logger = logging.getLogger(__name__)
    logger.info("Performing Bayesian estimates")
    logger.debug("Call: %s" % (" ".join(arg),))
    start_time = time.time()

    try:
        check_call(arg)
    except CalledProcessError as e:
        logger.error("The Bayesian statistics program failed to finish")
        raise e

    logger.debug(
        "Bayesian estimates completed in %f seconds" % (time.time() - start_time,)
    )
    return 0


def bool_to_int(boolean):
    """return 0 for False and 1 for True"""
    result = 1 if boolean else 0
    return result



import mapdamage

# from Martin Kircher, description of CIGAR operations
# O BAM Description
# M 0 alignment match (can be a sequence match or mismatch)
# I 1 insertion to the reference
# D 2 deletion from the reference
# N 3 skipped region from the reference
# S 4 soft clipping (clipped sequences present in SEQ)
# H 5 hard clipping (clipped sequences NOT present in SEQ)
# P 6 padding (silent deletion from padded reference)
# = 7 sequence match
# X 8 sequence mismatch


def get_coordinates(read):
    """return external coordinates of aligned read bases"""
    fivep = read.aend if read.is_reverse else read.pos
    threep = read.pos if read.is_reverse else read.aend

    return fivep, threep


def get_around(coord, chrom, reflengths, length, ref):
    """return reference sequences before and after the read
    check for extremities and return what is available"""
    coord_min = min(coord)
    coord_max = max(coord)

    pos_before = max(0, coord_min - length)
    pos_after = min(reflengths[chrom], coord_max + length)

    # Uppercased, to be sure that we don't compare A,C,G,T and a,c,g,t
    before = ref.fetch(chrom, pos_before, coord_min).upper()
    after = ref.fetch(chrom, coord_max, pos_after).upper()

    return before, after


def align(cigarlist, seq, ref):
    """insert gaps according to the cigar string
    deletion: gaps to be inserted into read sequences,
    insertions: gaps to be inserted into reference sequence"""
    lref = list(ref)
    for nbr, idx in parse_cigar(cigarlist, 1):
        lref[idx:idx] = ["-"] * nbr

    lread = list(seq)
    for nbr, idx in parse_cigar(cigarlist, 2):
        lread[idx:idx] = ["-"] * nbr

    return "".join(lread), "".join(lref)


def align_with_qual(cigarlist, seq, qual, threshold, ref):
    """insert gaps according to the cigar string
    deletion: gaps to be inserted into read sequences and qualities,
    insertions: gaps to be inserted into reference sequence"""
    lref = list(ref)
    for nbr, idx in parse_cigar(cigarlist, 1):
        lref[idx:idx] = ["-"] * nbr

    lread = list(seq)
    lqual = list(qual)
    for nbr, idx in parse_cigar(cigarlist, 2):
        lread[idx:idx] = ["-"] * nbr
        lqual[idx:idx] = ["-"] * nbr

    for idx, score in enumerate(lqual):
        # mask read and reference nucleotides (so not gaps) if below desired threshold
        if (ord(score) - 33) < threshold and lread[idx] != "-":
            lread[idx] = "N"
            lref[idx] = "N"

    return "".join(lread), "".join(lqual), "".join(lref)


def get_mis(read, seq, refseq, ref, length, tab, end):
    """count mismatches using aligned reference and read,
    must be redone since N in reference were randomly replaced by any bases"""
    std = "-" if read.is_reverse else "+"
    subtable = tab[ref][end][std]

    for i, nt_seq, nt_ref in zip(range(length), seq, refseq):
        if (nt_seq in "ACGT-") and (nt_ref in "ACGT-"):
            if nt_ref != "-":
                # record base composition in the reference, only A, C, G, T
                subtable[nt_ref][i] += 1

            # Most ref/seq pairs will be identical
            if nt_ref != nt_seq:
                mut = "%s>%s" % (nt_ref, nt_seq)
                subtable[mut][i] += 1


def parse_cigar(cigarlist, ope):
    """for a specific operation (mismatch, match, insertion, deletion... see above)
    return occurrences and index in the alignment"""
    tlength = 0
    coordinate = []
    # count matches, indels and mismatches
    oplist = (0, 1, 2, 7, 8)
    for operation, length in cigarlist:
        if operation == ope:
            coordinate.append([length, tlength])
        if operation in oplist:
            tlength += length
    return coordinate


def record_soft_clipping(read, tab, length):
    """record soft clipped bases at extremities"""

    def update_table(end, std, bases):
        for i in range(0, min(bases, length)):
            tab[end][std]["S"][i] += 1

    strand = "-" if read.is_reverse else "+"
    for nbases, idx in mapdamage.align.parse_cigar(read.cigar, 4):
        if idx == 0:
            # Soft-clipping at the left side of the alignment
            end = "3p" if read.is_reverse else "5p"
        else:
            # Soft-clipping at the right side of the alignment
            end = "5p" if read.is_reverse else "3p"

        update_table(end, strand, nbases)
